// Regulated Current PWM Pulse Testcase
// Ed Nisley - KE4ANU - November 2013

#include <stdio.h>

//----------
// Pin assignments

const byte PIN_READ_VBATT = 0;		// AI - battery voltage from divider
const byte PIN_READ_VGATE_A = 1;	// AI - MOSFET A gate voltage
const byte PIN_READ_VGATE_B = 2;	// AI - MOSFET B gate voltage
const byte PIN_READ_HALL = 3;		// AI - raw Hall sensor voltage
const byte PIN_READ_CURRENT = 4;	// AI - current sense amp
const byte PIN_READ_OFFSET = 5;		// AI - raw offset voltage

const byte PIN_SET_VGATE_A = 3;		// PWM - MOSFET A gate voltage
const byte PIN_SET_VGATE_B = 11;	// PWM - MOSFET B gate voltage
const byte PIN_SET_OFFSET = 9;		// PWM - VCC/2 bias offset voltage

const byte PIN_POWERON = 4;			// DO - high = enable power MOSFET
const byte PIN_STATUSLED = 10;		// DO - high = status LED
const byte PIN_SYNC = 12;			// DO - scope sync
const byte PIN_HEARTBEAT = 13;		// DO - high = standard Arduino LED

const byte PIN_POWER_BUTTON = 2;	// DI - low = button pressed
const byte PIN_MODE_A = 5;			// DI - low = mode A input active
const byte PIN_MODE_B = 6;			// DI - low = mode B input active

//----------
// Constants

const float NomLEDCurrent = 0.160;		// nominal LED current

const float Vcc = 4.96; 				// measured Arduino supply

const float VBattRatio = 3.75/7.50;		// measured division ratio for battery divider

const float VStep = Vcc/256;			// minimum PWM voltage increment = 5 V / 256
const float IDeadBand = 0.2*NomLEDCurrent;	// total width of LED current deadband

const float IGain = 0.195/1.809;		// measured LED current / Hall sense voltage

const byte PWM_Settle = 7;				// PWM settling time ms

// Timer prescaler = 1:1 for 32 kHz PWM
#define TCCRxB 0x01

#define MK_UL(fl,sc) ((unsigned long)((fl)*(sc)))
#define MK_U(fl,sc) ((unsigned int)((fl)*(sc)))
#define MK_I(fl,sc) ((int)((fl)*(sc)))

//----------
// Globals

float AVRef1V1;					// 1.1 V bandgap reference - calculated from Vcc

float VBatt;					// battery voltage - calculated from divider
float VGateSense;				// actual gate voltage - measured after PWM filter
float ILEDSense;				// LED current from Hall effect sensor

float VGateDrive;				// gate drive voltage

byte PWMHallOffset;				// zero-field Hall effect sensor bias

long unsigned long MillisNow, MillisThen;	// sampled millis() value

//-- Read AI channel
//      averages several readings to improve noise performance
//		returns value in volts assuming known VCC ref voltage

#define NUM_T_SAMPLES    10

float ReadAI(byte PinNum) {

  word RawAverage;

  digitalWrite(PIN_SYNC,HIGH);					// scope sync

  RawAverage = (word)analogRead(PinNum);		// prime the averaging pump

  for (int i=2; i <= NUM_T_SAMPLES; i++) {
    RawAverage += (word)analogRead(PinNum);
  }

  digitalWrite(PIN_SYNC,LOW);

  RawAverage /= NUM_T_SAMPLES;

  return Vcc * (float)RawAverage / 1024.0;

}

//-- Read & scale total current

float ReadCurrent(void) {
	return IGain * ReadAI(PIN_READ_CURRENT);
}

//-- Set PWM output

void SetPWMVoltage(byte PinNum,float PWMVolt) {

byte PWM;

  PWM = constrain((byte)(255.0 * PWMVolt / Vcc),0,255);

  analogWrite(PinNum,PWM);
  delay(PWM_Settle);

//  printf("PWM: %d\r\n",(unsigned)PWM);

}


//-- compute actual 1.1 V bandgap reference based on known VCC = AVcc (more or less)
//		adapted from http://code.google.com/p/tinkerit/wiki/SecretVoltmeter

float ReadBandGap(void) {

  word ADCBits;
  float VBandGap;

  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);	// select 1.1 V input
  delay(2); // Wait for Vref to settle

  ADCSRA |= _BV(ADSC);										// Convert
  while (bit_is_set(ADCSRA,ADSC));

  ADCBits = ADCL;
  ADCBits |= ADCH<<8;

  VBandGap = Vcc * (float)ADCBits / 1024.0;
  return VBandGap;
}

//-- null Hall sensor Vcc/2 offset
//   uses & sets global offset value
//   We can only sense positive current, so offset must be run upward from below

void FindSensorNull(void) {

#define MAX_OFFSET_OK 0.005
float Sense;

	digitalWrite(PIN_STATUSLED,HIGH);

//	printf("Nulling Hall sensor offset %d mV below %d mV",
//            MK_U(ReadAI(PIN_READ_HALL),1000.0),MK_U(MAX_OFFSET_OK,1000.0));

	analogWrite(PIN_SET_VGATE_A,0);			// force gate voltage = 0
	analogWrite(PIN_SET_VGATE_B,0);			//   ... so current sense output should = 0.0

//	printf("\r\n  Initial: %d PWM  %d mA ",
//		   PWMHallOffset,MK_U(ReadCurrent(),1000.0));

	while ((ReadCurrent() < MAX_OFFSET_OK) && PWMHallOffset) {	// force nonzero sense output
		PWMHallOffset -= 1;
		analogWrite(PIN_SET_OFFSET,PWMHallOffset);
//		printf(".");
		delay(PWM_Settle);
	}
	
//	printf("\r\n  Adjusted: PWM %d PWM  %d mA ",
//		   PWMHallOffset,MK_U(ReadCurrent(),1000.0));

	while (ReadCurrent() > MAX_OFFSET_OK) {
		PWMHallOffset += 1;
		analogWrite(PIN_SET_OFFSET,PWMHallOffset);
//		printf(".");
		delay(PWM_Settle);
	}

//	printf("\r\n  Final: %d PWM  %d mA\r\n",
//		   PWMHallOffset,MK_U(ReadCurrent(),1000.0));

	digitalWrite(PIN_STATUSLED,LOW);

}

//-- Helper routine for printf()

int s_putc(char c, FILE *t) {
  Serial.write(c);
}


//------------------
// Set things up

void setup() {
	
float AVRef1V1;

	pinMode(PIN_POWERON,OUTPUT);		// turn on the power!
	digitalWrite(PIN_POWERON,HIGH);		//   ... now the button can go low
	
	pinMode(PIN_HEARTBEAT,OUTPUT);
	digitalWrite(PIN_HEARTBEAT,HIGH);	// show we arrived

	pinMode(PIN_SYNC,OUTPUT);
	digitalWrite(PIN_SYNC,LOW);
	
	pinMode(PIN_STATUSLED,OUTPUT);
	digitalWrite(PIN_STATUSLED,LOW);

	TCCR1B = TCCRxB;					// set frequency for PWM 9 & 10
	TCCR2B = TCCRxB;					// set frequency for PWM 3 & 11

	analogWrite(PIN_SET_VGATE_A,0);	// force gate voltage = 0
	analogWrite(PIN_SET_VGATE_B,0);

	Serial.begin(57600);
	fdevopen(&s_putc,0);				// set up serial output for printf()

	printf("Hall effect Current Regulator - Pulse Tests\r\nEd Nisley - KE4ZNU - November 2013\r\n");

	printf("Assume Vcc: %d mV\r\n",MK_U(Vcc,1000.0));
	printf("Assume VBatt divider ratio: 0.%d\r\n",MK_U(VBattRatio,1000.0));

	AVRef1V1 = ReadBandGap();			// compute actual bandgap reference voltage
	printf("Bandgap reference voltage: %d mV\r\n",MK_U(AVRef1V1,1000.0));

	PWMHallOffset = 0;
	analogWrite(PIN_SET_OFFSET,PWMHallOffset);
	FindSensorNull();

	VBatt = ReadAI(PIN_READ_VBATT) / VBattRatio;
	printf("Battery voltage: %d mV\r\n",MK_U(VBatt,1000.0));

	printf("LED setpoint: %d mA\r\n",MK_U(NomLEDCurrent,1000.0));

	VGateDrive = 1.0;					// reasonable starting point
    ILEDSense = ReadCurrent();          // should be about zero

	MillisThen = millis();
}

//------------------
// Run the test loop

void loop() {

	if ((millis() - MillisThen) > 500) {
		MillisThen = millis();
		
		digitalWrite(PIN_HEARTBEAT,HIGH);
		
		if (ILEDSense < (NomLEDCurrent - IDeadBand/2.0)) {
			VGateDrive += VStep;
            printf("+");
		}
		else if (ILEDSense > (NomLEDCurrent + IDeadBand/2.0)) {
			VGateDrive -= VStep;
            printf("-");
		} else printf(" ");

		SetPWMVoltage(PIN_SET_VGATE_A,VGateDrive);

		VGateSense = ReadAI(PIN_READ_VGATE_A);
		ILEDSense = ReadCurrent();

        printf("Gate drive: %d mV  ",MK_U(VGateDrive,1000.0));
		printf("sense: %d mV  ",MK_U(VGateSense,1000.0));
		printf("LED: %d mA ",MK_U(ILEDSense,1000.0));
		
		delay(20 - PWM_Settle);
		SetPWMVoltage(PIN_SET_VGATE_A,0.0);
		
		delay(75);
		SetPWMVoltage(PIN_SET_VGATE_A,VGateDrive);
		ILEDSense = ReadCurrent();
		printf("%d mA\r\n",MK_U(ILEDSense,1000.0));

		delay(20 - PWM_Settle);
		SetPWMVoltage(PIN_SET_VGATE_A,0.0);
		
		FindSensorNull();
		
		digitalWrite(PIN_HEARTBEAT,LOW);
	}
}

